import sqlalchemy as sql
from sqlalchemy.ext.declarative import declarative_base

from . import config as C
from . import wrapper


engine = sql.create_engine(C.URL, **C.CREATE_ENGINE)
Base = declarative_base(bind=engine)
Session = sql.orm.sessionmaker(bind=engine)


class StockExchange(Base):
    __tablename__ = "stock_exchange"

    id = sql.Column(sql.Integer, primary_key=True)
    name = sql.Column(sql.Text, nullable=False)
    company = sql.orm.relationship(
        "Company",
        backref="stock_exchange",
        uselist=False
    )


class DayInfo(Base):
    __tablename__ = "day_info"
    __table_args__ = (sql.UniqueConstraint('date', 'company_id'), )

    id = sql.Column(sql.Integer, primary_key=True)
    high = sql.Column(sql.Float, nullable=False)
    low = sql.Column(sql.Float, nullable=False)
    opening = sql.Column(sql.Float, nullable=False)
    closing = sql.Column(sql.Float, nullable=False)
    date = sql.Column(sql.Date, nullable=False)
    volume = sql.Column(sql.Integer, nullable=False)
    company_id = sql.Column(
        sql.Integer,
        sql.ForeignKey("company.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False
    )
    company = sql.orm.relation(
        'Company',
        order_by='Company.id',
        uselist=False,
        backref='day_info_list'
    )

    @property
    def w(self):
        return wrapper.DayInfo(self)


class SplitStockDate(Base):

    __tablename__ = "split_stock_date"
    __table_args__ = (sql.UniqueConstraint("date", "company_id"), )

    id = sql.Column(sql.Integer, primary_key=True)
    from_number = sql.Column(sql.Integer, nullable=False)
    to_number = sql.Column(sql.Integer, nullable=False)
    date = sql.Column(sql.Date, nullable=False)
    company_id = sql.Column(
        sql.Integer,
        sql.ForeignKey("company.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False
    )
    company = sql.orm.relation(
        "Company",
        order_by="Company.id",
        uselist=False,
        backref="split_stock_dates"
    )


class Company(Base):
    __tablename__ = "company"
    __table_args__ = (sql.UniqueConstraint("code"), )

    id = sql.Column(sql.Integer, primary_key=True)
    name = sql.Column(sql.Text, nullable=False)
    description = sql.Column(sql.Text, nullable=True)
    code = sql.Column(sql.Integer)
    category17 = sql.Column(sql.Integer)
    category33 = sql.Column(sql.Integer)
    scale = sql.Column(sql.Integer, nullable=False)
    stock_exchange_id = sql.Column(
        sql.Integer,
        sql.ForeignKey('stock_exchange.id', onupdate="CASCADE", ondelete="CASCADE"),
        nullable=True,
    )

    @property
    def w(self):
        return wrapper.Company(self)

    @sql.orm.validates('scale')
    def validate_scale(self, key, scale):
        if isinstance(scale, int):
            return scale
        elif isinstance(scale, float):
            return int(scale)
        else:
            raise ValueError("%s is not allowed for scale." % scale)


class CurrentValue(Base):
    __tablename__ = "current_value"

    id = sql.Column(sql.Integer, primary_key=True)
    value = sql.Column(sql.Integer)
    datetime = sql.Column(sql.DateTime, nullable=False)



from flask import Flask, render_template, request, jsonify, abort

from stock import query


app = Flask(__name__)


@app.route("/", methods=["GET"])
@app.route("/company/", methods=["GET"])
def index():
    return render_template('index.html', **{"company_list": query.Company.query()})


@app.route("/company/<int:id>", methods=["GET"])
def company(id):
    company = query.Company.first(id=id)
    if not company:
        abort(404)
    return render_template('company.html', **{"company": company})


@app.route("/api/<int:company_id>", methods=["GET"])
def api(company_id):
    company = query.Company.first(id=company_id)
    if not company:
        abort(404)
    q = query.DayInfo.get(company_id)
    return jsonify({
        "company": company.w.to_dict(),
        "series": [
            {"name": "closing", "data": q.closing()},
            {"name": "rolling_mean25", "data": q.rolling_mean(period=25)},
            {"name": "rolling_mean5", "data": q.rolling_mean(period=5)},
        ]
    })



# coding: utf-8
import os
import path
from collections import OrderedDict

ROOTDIR = path.Path(__file__).parent.parent
HERE = os.path.dirname(__file__)

DEBUG = True

PORT = int(os.environ.get('PORT', 5000))

# URL = "sqlite:///:memory:"
# URL = "mysql+mysqlconnector://root@localhost/stock?charset=utf8"
DEFAULT_URL = "mysql+pymysql://root@localhost/stock?charset=utf8"
URL = os.environ.get("DATABASE_URL", DEFAULT_URL)

# sqlalchemy
CREATE_ENGINE = {
    "encoding": 'utf-8',
    "echo": DEBUG,
}

STATIC_DIR = os.path.join(ROOTDIR, "static")

GRAPH_DIR = os.path.join(STATIC_DIR, "company")

FORMAT = {
    "image_dir": os.path.join(GRAPH_DIR, "{code}"),
    "month": "{code}/{year}_{month}",
}


# frequently changed
COMPANY_XLS_URL = "http://www.jpx.co.jp/markets/statistics-equities/misc/tvdivq0000001vg2-att/data_j.xls"
EXCEL_COMPANY_HEADER = OrderedDict([
    ("date", u'日付'),
    ("code", u'コード'),
    ("name", u'銘柄名'),
    ("_item", u'市場・商品区分'),
    ("category33", u'33業種コード'),
    ("label33", u'33業種区分'),
    ("category17", u'17業種コード'),
    ("label17", u'17業種区分'),
    ("scale", u'規模コード'),
    ("label_scale", u'規模区分'),
])
SHEET_NAME = "Sheet1"

# 今日から株価を取得する期間
DEFAULT_DAYS_PERIOD = 3 * 30



import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from s import util


class Graph(object):

    @classmethod
    def month(cls, company, year, month):
        date = util.Date(year, month)
        dayinfos = company.range(date.first, date.last)
        manager = util.DayInfoManager(dayinfos)

        fig = plt.figure()
        loc = mdates.DayLocator()
        ax = fig.add_subplot(111)

        # labels = ax.get_xticklabels()
        # plt.setp(labels, rotation=30, fontsize=10)
        plt.xticks(rotation=30)

        ax.xaxis.set_major_locator(loc)
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
        ax.grid(True)
        #ax.plot(list(zip(*manager.sequence_high)))
        xy = list(zip(*manager.sequence_high))
        fig.autofmt_xdate()
        ax.plot(*xy)

        filepath = util.graph_month_dir(company.code, year, month)
        plt.savefig(filepath)



import datetime

import sqlalchemy as sql
import pandas as pd


class Query(object):

    def __init__(self, q):
        self.q = q

    def __iter__(self):
        return iter(self.q)

    def __getattr__(self, name):
        attr = getattr(self.q, name)
        if callable(attr):
            def f(*args, **kw):
                r = attr(*args, **kw)
                if isinstance(r, sql.orm.query.Query):
                    return self.__class__(r)
                else:
                    return r
            return f
        else:
            return attr


class DayInfoQuery(Query):

    def df(self):
        return pd.DataFrame(
            [(info.w.js_datetime, info.w.closing) for info in self],
            columns=["date", "closing"]
        )

    def rolling_mean(self, period=30):
        df = self.df()
        mean = pd.rolling_mean(df.closing, period)
        return list([a, b] for a, b in zip(df.date.values.tolist(), mean.values.tolist())
                    if not pd.isnull(b))

    def closing(self):
        df = self.df()
        return list(zip(df.date.values.tolist(),
                        df.closing.values.tolist()))

    def to_series(self, type="closing"):
        return [(info.w.js_datetime, info.w.closing) for info in self]


class Wrapper(object):

    def __init__(self, ins):
        from . import models  # Not cyclic import
        assert isinstance(ins, getattr(models, self.__class__.__name__))
        self.ins = ins

    # company_idみたいに、定義されていない属性はinsから取得
    def __getattr__(self, name):
        return getattr(self.ins, name)

    def to_dict(self):
        assert getattr(self, "dict_keys", False), "Must need a dict_keys attribute"
        return {k: getattr(self, k) for k in self.dict_keys}


class Company(Wrapper):
    dict_keys = ["id", "name", "code"]

    def fix_data_frame(self):
        data_records = []
        for day_info in self.day_info_list:
            data = {
                "high": day_info.fix_high(),
                "low": day_info.fix_low(),
                "opening": day_info.fix_opening(),
                "closing": day_info.fix_closing(),
                "date":day_info.js_datetime}
            data_records.append(data)
        return pd.DataFrame.from_records(data_records)


class DayInfo(Wrapper):

    dict_keys = ["high", "low", "opening", "closing", "company_id", "js_datetime"]

    def __str__(self):
        return """\
cid  : {company_id}
open : {opening}
high : {high}
low  : {low}
close: {closing}
""".format(**self.to_dict())

    def _fix_value(self, value):
        for date in self.ins.company.split_stock_dates:
            if self.date < date.date:
                value *= date.from_number / float(date.to_number)
        return value

    @property
    def high(self):
        return self._fix_value(self.ins.high)

    @property
    def low(self):
        return self._fix_value(self.ins.low)

    @property
    def opening(self):
        return self._fix_value(self.ins.opening)

    @property
    def closing(self):
        return self._fix_value(self.ins.closing)

    @property
    def js_datetime(self):
        japan = self.ins.date + datetime.timedelta(hours=9)
        return int(japan.strftime("%s")) * 1000



import os
import time
import calendar
import datetime

from dateutil.relativedelta import relativedelta
import sqlalchemy as sql

from . import config as C


class DateRange(object):

    def __init__(self, start=None, end=None):
        if isinstance(end, str):
            end = str2date(start)
        if isinstance(start, str):
            start = str2date(start)
        if end is None:
            end = datetime.date.today()
        if start is None:
            start = end - relativedelta(days=C.DEFAULT_DAYS_PERIOD)
        self.end = end
        self.start = start

    def query(self, date_col):
        start, end = self.start, self.end
        if start is None:
            return date_col <= end
        elif end is None:
            return start <= date_col
        else:
            return sql.and_(start <= date_col, date_col <= end)

    def to_dict(self):
        return {"start": str(self.start), "end": str(self.end)}

    def to_short_dict(self):
        return {
            "sy": self.start.year,
            "sm": self.start.month,
            "sd": self.start.day,
            "ey": self.end.year,
            "em": self.end.month,
            "ed": self.end.day,
        }


class Date(object):

    def __init__(self, year, month):
        self.first = datetime.date(year, month, 1)
        self.last = self.first + relativedelta(months=1, days=-1)

    def to_days(self):
        days = []
        if self.first < self.last:
            i = self.first
            while True:
                days.append(i)
                i += relativedelta(days=1)
                if i == self.last:
                    return days
        else:
            return []

    def __str__(self):
        return "from {first} to {last}".format(**self)


def graph_month_dir(code, year, month, ext=".png"):
    name = C.FORMAT["month"].format(**locals())
    dir_ = C.FORMAT["image_dir"].format(code=code)
    if not os.path.isdir(dir_):
        os.makedirs(dir_)
    filepath = os.path.join(C.GRAPH_DIR, name + ext)
    return filepath


def last_date():
    """株の最後の日を返す"""
    today = datetime.date.today()
    weekday = today.weekday()
    if weekday in [calendar.SUNDAY, calendar.SATURDAY]:
        date = today + relativedelta.relativedelta(weekday=relativedelta.FR(-1))
    else:
        date = today - relativedelta(days=1)
    return date


def dict_inverse(dct):
    return {v: k for k, v in dct.items()}


def multiple_decorator(funcs):
    def wrap(g):
        for f in funcs:
            g = f(g)
    return wrap


def str2date(datestr):
    if datestr:
        t = time.strptime(datestr, "%Y-%m-%d")
        return datetime.date.fromtimestamp(time.mktime(t))



from logging import getLogger

import sqlalchemy as sql
import pandas as pd

from . import models
from . import util
from . import wrapper

logger = getLogger(__name__)


class Query(object):
    model = None

    @classmethod
    def session(cls):
        return models.Session()

    @classmethod
    def query(cls, session=None):
        if session is None:
            session = cls.session()
        return session.query(cls.model)

    @classmethod
    def one(cls, id, session=None):
        return cls.query(session).filter_by(id=id).one()

    @classmethod
    def first(cls, session=None, **kw):
        return cls.query(session).filter_by(**kw).first()


class DayInfo(Query):
    model = models.DayInfo

    @classmethod
    def get(cls, company_id, start=None, end=None):
        q = cls.query().filter_by(company_id=company_id)
        q = wrapper.DayInfoQuery(q)
        q = q.filter(util.DateRange(start, end).query(cls.model.date))
        q = q.order_by("date")
        return q

    @classmethod
    def set(cls, company_id, start=None, end=None, each=False, ignore=False, last_date=None):
        # TODO: use with statement
        session = cls.session()
        c = Company.first(id=company_id, last_date=last_date, session=session)
        if not c:
            logger.info("SKIP: company(id=%s)" % company_id)
            session.close()
            return

        from stock.scrape import YahooJapan
        scraper = YahooJapan()
        history = scraper.history(c.code, start, end)
        for d in history:
            d["company_id"] = company_id
            session.add(models.DayInfo(**d))
            if not each:
                continue
            try:
                session.commit()
            except sql.exc.IntegrityError:
                session.rollback()
        if not each:
            try:
                session.commit()
            except sql.exc.IntegrityError as e:
                if not ignore:
                    raise e
        session.close()

    @classmethod
    def sets(cls, min_id=1, max_id=None, start=None, end=None,
             each=False, ignore=False, last_date=None):
        if max_id is None:
            max_id = Company.max_id()
        for id in range(min_id, max_id + 1):
            cls.set(id, each=each, ignore=True, last_date=last_date)


class Company(Query):
    model = models.Company

    @classmethod
    def first(cls, session=None, last_date=None, **kw):
        q = cls.query(session).filter_by(**kw)
        if last_date:
            q = q.filter(sql.not_(models.Company.day_info_list.any(date=last_date)))
        return q.first()

    @classmethod
    def max_id(cls):
        q = cls.query()
        q = q.order_by("-id")
        c = q.first()
        return c.id if c else 0

    @classmethod
    def is_updated(cls, id):
        """会社の最新情報に更新されていればTrue"""
        last = util.last_day()
        q = cls.query()
        q = q.filter_by(id=id)
        return q.count() > 0


def go_down_rolling_mean():
    """長期移動平均線を下回っている株を表示"""
    low_cost_company_list = []
    for c in Company.query():
        df = c.fix_data_frame()
        mean = pd.rolling_mean(df.closing, 90)
        cmp = mean.tail(1) > df.closing.tail(1)
        if bool(cmp):
            low_cost_company_list.append(c)
    return low_cost_company_list



from setuptools import setup, find_packages


with open('requirements.txt') as f:
    install_requires = f.read().splitlines()


setup(
    name='stock',
    version='0.3',
    author='Hiroyuki Ishii',
    py_modules=['stock'],
    packages=find_packages(),
    description='A command line tool for stock markets',
    keywords=['stock'],
    install_requires=install_requires,
    entry_points='''
        [console_scripts]
        pystock=stock.cmd:cli
    ''',
)



import io
import logging

import click
import requests

from stock import models
from stock import query
from stock import scrape
from stock import util
from stock import config as C

from .import_company import Reader


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def mkdate(ctx, param, datestr):
    return util.str2date(datestr)


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


@click.group(cls=AliasedGroup, invoke_without_command=True)
def cli():
    ctx = click.get_current_context()
    if ctx.invoked_subcommand is None:
        pass
        # ctx.invoke(default)


@cli.group()
def store():
    pass


@click.option("--start", callback=mkdate)
@click.option("--end", callback=mkdate)
@click.option("--code")
@click.option("--each", is_flag=True, default=True)
@store.command()
def history(code, start, end, each):
    query.DayInfo.set(code, start, end, each=each)


@store.command(help="""\
You can download the xls at \
http://www.jpx.co.jp/markets/statistics-equities/misc/01.html
or directly {url}
""".format(url=C.COMPANY_XLS_URL))
def company():
    resp = requests.get(C.COMPANY_XLS_URL)
    if resp.ok:
        xls = resp.content
        Reader(filepath=io.BytesIO(xls)).store()
    else:
        click.echo("Can't get %s" % C.COMPANY_XLS_URL)


@cli.group()
def db():
    pass


@db.command(help="Create new all tables")
def create():
    models.Base.metadata.create_all()


@cli.command(help="Setup")
@click.option("--all", is_flag=True, default=False)
def setup(all):
    click.echo("Start setup ...")
    ctx = click.get_current_context()
    for cmd in [create, company]:
        ctx.invoke(cmd)


@cli.command(help="show info")
@click.argument('code', type=int)
def show(code):
    for info in query.DayInfo.get(code):
        print(info.w)


@click.option("--port", default=C.PORT, type=int)
@click.option("--debug", default=C.DEBUG, is_flag=True)
@cli.command(help="Start server")
def serve(port, debug):
    from stock.server import app
    app.run(port=port, debug=debug)


@cli.group()
def scrape_():
    pass

_decorator = util.multiple_decorator([
    click.option("--scraper", default=scrape.YahooJapan),
    click.option("--start"),
    click.option("--end"),
    click.argument('code', type=int)
])


@_decorator
@scrape_.command(name="history")
def scrape_history(code, start, end, scraper):
    history = scraper.history(code, start, end)
    for day_info in history:
        click.echo(day_info)


@_decorator
@scrape_.command(name="split")
def split_stock_date(code, start, end, scraper):
    history = scraper.split_stock_date(code, start, end)
    for day_info in history:
        click.echo(day_info)


@click.option("--scraper", default=scrape.YahooJapan)
@click.argument('code', type=int)
@scrape_.command(name="day_info")
def day_info(code, scraper):
    day_info = scraper.day_info(code)
    click.echo(day_info)


@click.option("--scraper", default=scrape.YahooJapan)
@click.argument('code', type=int)
@scrape_.command(name="value")
def current_value(code, scraper):
    value = scraper.current_value(code)
    click.echo(value)


@click.option("--max-id", type=int)
@click.option("--min-id", type=int, default=1)
@click.option("--each", is_flag=True, default=False)
@click.option("--last-date", callback=mkdate)
@cli.command(help="update day info")
def update(min_id, max_id, each, last_date):
    query.DayInfo.sets(min_id=min_id, max_id=max_id,
                       each=each, ignore=True, last_date=last_date)
